@@ -1,5 +1,4 @@
-from __future__ import unicode_literals
-
+import uuid
 from operator import attrgetter
 
 from django import VERSION
@@ -15,16 +14,19 @@ from django.db.models.fields.related import (
     lazy_related_operation,
 )
 from django.db.models.query_utils import PathInfo
-from django.utils import six
 from django.utils.text import capfirst
-from django.utils.translation import ugettext_lazy as _
+from django.utils.translation import gettext_lazy as _
 
 from taggit.forms import TagField
-from taggit.models import CommonGenericTaggedItemBase, TaggedItem
+from taggit.models import (
+    CommonGenericTaggedItemBase,
+    GenericUUIDTaggedItemBase,
+    TaggedItem,
+)
 from taggit.utils import require_instance_manager
 
 
-class ExtraJoinRestriction(object):
+class ExtraJoinRestriction:
     """
     An extra restriction used for contenttype restriction in joins.
     """
@@ -39,12 +41,10 @@ class ExtraJoinRestriction(object):
     def as_sql(self, compiler, connection):
         qn = compiler.quote_name_unless_alias
         if len(self.content_types) == 1:
-            extra_where = "%s.%s = %%s" % (qn(self.alias), qn(self.col))
+            extra_where = "{}.{} = %s".format(qn(self.alias), qn(self.col))
         else:
-            extra_where = "%s.%s IN (%s)" % (
-                qn(self.alias),
-                qn(self.col),
-                ",".join(["%s"] * len(self.content_types)),
+            extra_where = "{}.{} IN ({})".format(
+                qn(self.alias), qn(self.col), ",".join(["%s"] * len(self.content_types))
             )
         return extra_where, self.content_types
 
@@ -52,12 +52,14 @@ class ExtraJoinRestriction(object):
         self.alias = change_map.get(self.alias, self.alias)
 
     def clone(self):
-        return self.__class__(self.alias, self.col, self.content_types[:])
+        return type(self)(self.alias, self.col, self.content_types[:])
 
 
 class _TaggableManager(models.Manager):
+    # TODO investigate whether we can use a RelatedManager instead of all this stuff
+    # to take advantage of all the Django goodness
     def __init__(self, through, model, instance, prefetch_cache_name):
-        super(_TaggableManager, self).__init__()
+        super().__init__()
         self.through = through
         self.model = model
         self.instance = instance
@@ -78,7 +80,7 @@ class _TaggableManager(models.Manager):
             raise ValueError("Custom queryset can't be used for this lookup.")
 
         instance = instances[0]
-        db = self._db or router.db_for_read(instance.__class__, instance=instance)
+        db = self._db or router.db_for_read(type(instance), instance=instance)
 
         fieldname = (
             "object_id"
@@ -101,14 +103,30 @@ class _TaggableManager(models.Manager):
             .using(db)
             .extra(
                 select={
-                    "_prefetch_related_val": "%s.%s" % (qn(join_table), qn(source_col))
+                    "_prefetch_related_val": "{}.{}".format(
+                        qn(join_table), qn(source_col)
+                    )
                 }
             )
         )
+
+        if issubclass(self.through, GenericUUIDTaggedItemBase):
+
+            def uuid_rel_obj_attr(v):
+                value = attrgetter("_prefetch_related_val")(v)
+                if value is not None and not isinstance(value, uuid.UUID):
+                    input_form = "int" if isinstance(value, int) else "hex"
+                    value = uuid.UUID(**{input_form: value})
+                return value
+
+            rel_obj_attr = uuid_rel_obj_attr
+        else:
+            rel_obj_attr = attrgetter("_prefetch_related_val")
+
         if VERSION < (2, 0):
             return (
                 qs,
-                attrgetter("_prefetch_related_val"),
+                rel_obj_attr,
                 lambda obj: obj._get_pk_val(),
                 False,
                 self.prefetch_cache_name,
@@ -116,7 +134,7 @@ class _TaggableManager(models.Manager):
         else:
             return (
                 qs,
-                attrgetter("_prefetch_related_val"),
+                rel_obj_attr,
                 lambda obj: obj._get_pk_val(),
                 False,
                 self.prefetch_cache_name,
@@ -127,10 +145,13 @@ class _TaggableManager(models.Manager):
         return self.through.lookup_kwargs(self.instance)
 
     @require_instance_manager
-    def add(self, *tags):
+    def add(self, *tags, through_defaults=None, tag_kwargs=None, **kwargs):
+
+        if tag_kwargs is None:
+            tag_kwargs = {}
         db = router.db_for_write(self.through, instance=self.instance)
 
-        tag_objs = self._to_tag_model_instances(tags)
+        tag_objs = self._to_tag_model_instances(tags, tag_kwargs)
         new_ids = {t.pk for t in tag_objs}
 
         # NOTE: can we hardcode 'tag_id' here or should the column name be got
@@ -155,7 +176,7 @@ class _TaggableManager(models.Manager):
 
         for tag in tag_objs:
             self.through._default_manager.using(db).get_or_create(
-                tag=tag, **self._lookup_kwargs()
+                tag=tag, **self._lookup_kwargs(), defaults=through_defaults
             )
 
         signals.m2m_changed.send(
@@ -168,7 +189,7 @@ class _TaggableManager(models.Manager):
             using=db,
         )
 
-    def _to_tag_model_instances(self, tags):
+    def _to_tag_model_instances(self, tags, tag_kwargs):
         """
         Takes an iterable containing either strings, tag objects, or a mixture
         of both and returns set of tag objects.
@@ -181,11 +202,11 @@ class _TaggableManager(models.Manager):
         for t in tags:
             if isinstance(t, self.through.tag_model()):
                 tag_objs.add(t)
-            elif isinstance(t, six.string_types):
+            elif isinstance(t, str):
                 str_tags.add(t)
             else:
                 raise ValueError(
-                    "Cannot add {0} ({1}). Expected {2} or str.".format(
+                    "Cannot add {} ({}). Expected {} or str.".format(
                         t, type(t), type(self.through.tag_model())
                     )
                 )
@@ -208,19 +229,19 @@ class _TaggableManager(models.Manager):
         else:
             # If str_tags has 0 elements Django actually optimizes that to not
             # do a query.  Malcolm is very smart.
-            existing = manager.filter(name__in=str_tags)
-            tags_to_create = str_tags - {t.name for t in existing}
+            existing = manager.filter(name__in=str_tags, **tag_kwargs)
+
+            tags_to_create = str_tags - set(t.name for t in existing)
 
         tag_objs.update(existing)
 
         for new_tag in tags_to_create:
             if case_insensitive:
-                tag, created = manager.get_or_create(
-                    name__iexact=new_tag, defaults={"name": new_tag}
-                )
+                lookup = {"name__iexact": new_tag, **tag_kwargs}
             else:
-                tag, created = manager.get_or_create(name=new_tag)
+                lookup = {"name": new_tag, **tag_kwargs}
 
+            tag, create = manager.get_or_create(**lookup, defaults={"name": new_tag})
             tag_objs.add(tag)
 
         return tag_objs
@@ -234,22 +255,27 @@ class _TaggableManager(models.Manager):
         return self.get_queryset().values_list("slug", flat=True)
 
     @require_instance_manager
-    def set(self, *tags, **kwargs):
+    def set(self, *tags, through_defaults=None, **kwargs):
         """
         Set the object's tags to the given n tags. If the clear kwarg is True
         then all existing tags are removed (using `.clear()`) and the new tags
         added. Otherwise, only those tags that are not present in the args are
         removed and any new tags added.
+
+        Any kwarg apart from 'clear' will be passed when adding tags.
+
         """
         db = router.db_for_write(self.through, instance=self.instance)
+
         clear = kwargs.pop("clear", False)
+        tag_kwargs = kwargs.pop("tag_kwargs", {})
 
         if clear:
             self.clear()
-            self.add(*tags)
+            self.add(*tags, **kwargs)
         else:
             # make sure we're working with a collection of a uniform type
-            objs = self._to_tag_model_instances(tags)
+            objs = self._to_tag_model_instances(tags, tag_kwargs)
 
             # get the existing tag strings
             old_tag_strs = set(
@@ -266,7 +292,7 @@ class _TaggableManager(models.Manager):
                     new_objs.append(obj)
 
             self.remove(*old_tag_strs)
-            self.add(*new_objs)
+            self.add(*new_objs, through_defaults=through_defaults, **kwargs)
 
     @require_instance_manager
     def remove(self, *tags):
@@ -344,7 +370,7 @@ class _TaggableManager(models.Manager):
     def similar_objects(self):
         lookup_kwargs = self._lookup_kwargs()
         lookup_keys = sorted(lookup_kwargs)
-        qs = self.through.objects.values(*six.iterkeys(lookup_kwargs))
+        qs = self.through.objects.values(*lookup_kwargs.keys())
         qs = qs.annotate(n=models.Count("pk"))
         qs = qs.exclude(**lookup_kwargs)
         qs = qs.filter(tag__in=self.all())
@@ -364,8 +390,9 @@ class _TaggableManager(models.Manager):
                     % remote_field.field_name: [r["content_object"] for r in qs]
                 }
             )
+            actual_remote_field_name = f.target_field.get_attname()
             for obj in objs:
-                items[(getattr(obj, remote_field.field_name),)] = obj
+                items[(getattr(obj, actual_remote_field_name),)] = obj
         else:
             preload = {}
             for result in qs:
@@ -408,7 +435,7 @@ class TaggableManager(RelatedField):
 
         rel = ManyToManyRel(self, to, related_name=related_name, through=self.through)
 
-        super(TaggableManager, self).__init__(
+        super().__init__(
             verbose_name=verbose_name,
             help_text=help_text,
             blank=blank,
@@ -426,40 +453,37 @@ class TaggableManager(RelatedField):
                 "%s objects need to have a primary key value "
                 "before you can access their tags." % model.__name__
             )
-        manager = self.manager(
+        return self.manager(
             through=self.through,
             model=model,
             instance=instance,
             prefetch_cache_name=self.name,
         )
-        return manager
 
     def deconstruct(self):
         """
         Deconstruct the object, used with migrations.
         """
-        name, path, args, kwargs = super(TaggableManager, self).deconstruct()
+        name, path, args, kwargs = super().deconstruct()
         # Remove forced kwargs.
         for kwarg in ("serialize", "null"):
             del kwargs[kwarg]
         # Add arguments related to relations.
         # Ref: https://github.com/jazzband/django-taggit/issues/206#issuecomment-37578676
         rel = self.remote_field
-        if isinstance(rel.through, six.string_types):
+        if isinstance(rel.through, str):
             kwargs["through"] = rel.through
         elif not rel.through._meta.auto_created:
-            kwargs["through"] = "%s.%s" % (
-                rel.through._meta.app_label,
-                rel.through._meta.object_name,
+            kwargs["through"] = "{}.{}".format(
+                rel.through._meta.app_label, rel.through._meta.object_name
             )
 
         related_model = rel.model
-        if isinstance(related_model, six.string_types):
+        if isinstance(related_model, str):
             kwargs["to"] = related_model
         else:
-            kwargs["to"] = "%s.%s" % (
-                related_model._meta.app_label,
-                related_model._meta.object_name,
+            kwargs["to"] = "{}.{}".format(
+                related_model._meta.app_label, related_model._meta.object_name
             )
 
         return name, path, args, kwargs
@@ -472,7 +496,7 @@ class TaggableManager(RelatedField):
         cls._meta.add_field(self)
         setattr(cls, name, self)
         if not cls._meta.abstract:
-            if isinstance(self.remote_field.model, six.string_types):
+            if isinstance(self.remote_field.model, str):
 
                 def resolve_related_class(cls, model, field):
                     field.remote_field.model = model
@@ -480,7 +504,7 @@ class TaggableManager(RelatedField):
                 lazy_related_operation(
                     resolve_related_class, cls, self.remote_field.model, field=self
                 )
-            if isinstance(self.through, six.string_types):
+            if isinstance(self.through, str):
 
                 def resolve_related_class(cls, model, field):
                     self.through = model
@@ -531,13 +555,13 @@ class TaggableManager(RelatedField):
         defaults.update(kwargs)
         return form_class(**defaults)
 
-    def value_from_object(self, instance):
-        if instance.pk:
-            return self.through.objects.filter(**self.through.lookup_kwargs(instance))
-        return self.through.objects.none()
-
-    def related_query_name(self):
-        return self.model._meta.model_name
+    def value_from_object(self, obj):
+        if obj.pk is None:
+            return []
+        qs = self.through.objects.select_related("tag").filter(
+            **self.through.lookup_kwargs(obj)
+        )
+        return [ti.tag for ti in qs]
 
     def m2m_reverse_name(self):
         return self.through._meta.get_field("tag").column
@@ -553,7 +577,7 @@ class TaggableManager(RelatedField):
 
     def m2m_column_name(self):
         if self.use_gfk:
-            return self.through._meta.virtual_fields[0].fk_field
+            return self.through._meta.private_fields[0].fk_field
         return self.through._meta.get_field("content_object").column
 
     def m2m_db_table(self):
