@@ -1,15 +1,13 @@
-from __future__ import unicode_literals
-
 import uuid
 
 from django.db import models
-from django.utils.encoding import python_2_unicode_compatible
 
 from taggit.managers import TaggableManager
 from taggit.models import (
     CommonGenericTaggedItemBase,
     GenericTaggedItemBase,
     GenericUUIDTaggedItemBase,
+    ItemBase,
     Tag,
     TagBase,
     TaggedItem,
@@ -41,7 +39,15 @@ class MultipleTagsGFK(models.Model):
     tags2 = TaggableManager(through=ThroughGFK, related_name="tagsgfk2")
 
 
-@python_2_unicode_compatible
+class BlankTagModel(models.Model):
+    name = models.CharField(max_length=50)
+
+    tags = TaggableManager(blank=True)
+
+    def __str__(self):
+        return self.name
+
+
 class Food(models.Model):
     name = models.CharField(max_length=50)
 
@@ -51,7 +57,29 @@ class Food(models.Model):
         return self.name
 
 
-@python_2_unicode_compatible
+class BaseFood(models.Model):
+    name = models.CharField(max_length=50)
+
+    def __str__(self):
+        return self.name
+
+
+class MultiInheritanceLazyResolutionFoodTag(TaggedItemBase):
+    content_object = models.ForeignKey(
+        "MultiInheritanceFood", related_name="tagged_items", on_delete=models.CASCADE
+    )
+
+    class Meta:
+        unique_together = [["content_object", "tag"]]
+
+
+class MultiInheritanceFood(BaseFood):
+    tags = TaggableManager(through=MultiInheritanceLazyResolutionFoodTag)
+
+    def __str__(self):
+        return self.name
+
+
 class Pet(models.Model):
     name = models.CharField(max_length=50)
 
@@ -71,12 +99,14 @@ class HousePet(Pet):
 class TaggedFood(TaggedItemBase):
     content_object = models.ForeignKey("DirectFood", on_delete=models.CASCADE)
 
+    class Meta:
+        unique_together = [["content_object", "tag"]]
+
 
 class TaggedPet(TaggedItemBase):
     content_object = models.ForeignKey("DirectPet", on_delete=models.CASCADE)
 
 
-@python_2_unicode_compatible
 class DirectFood(models.Model):
     name = models.CharField(max_length=50)
 
@@ -86,7 +116,6 @@ class DirectFood(models.Model):
         return self.name
 
 
-@python_2_unicode_compatible
 class DirectPet(models.Model):
     name = models.CharField(max_length=50)
 
@@ -100,18 +129,75 @@ class DirectHousePet(DirectPet):
     trained = models.BooleanField(default=False)
 
 
+# Test direct-tagging with custom through model and custom tag
+
+
+class TrackedTag(TagBase):
+    created_by = models.CharField(max_length=50)
+    created_dt = models.DateTimeField(auto_now_add=True)
+    description = models.TextField(blank=True, max_length=255, null=True)
+
+
+class TaggedTrackedFood(ItemBase):
+    content_object = models.ForeignKey("DirectTrackedFood", on_delete=models.CASCADE)
+    tag = models.ForeignKey(
+        TrackedTag, on_delete=models.CASCADE, related_name="%(class)s_items"
+    )
+    created_by = models.CharField(max_length=50)
+    created_dt = models.DateTimeField(auto_now_add=True)
+
+    class Meta:
+        unique_together = ["content_object", "tag"]
+
+
+class TaggedTrackedPet(ItemBase):
+    content_object = models.ForeignKey("DirectTrackedPet", on_delete=models.CASCADE)
+    tag = models.ForeignKey(
+        TrackedTag, on_delete=models.CASCADE, related_name="%(class)s_items"
+    )
+    created_by = models.CharField(max_length=50)
+    created_dt = models.DateTimeField(auto_now_add=True)
+
+
+class DirectTrackedFood(models.Model):
+    name = models.CharField(max_length=50)
+
+    tags = TaggableManager(through=TaggedTrackedFood)
+
+    def __str__(self):
+        return self.name
+
+
+class DirectTrackedPet(models.Model):
+    name = models.CharField(max_length=50)
+
+    tags = TaggableManager(through=TaggedTrackedPet)
+
+    def __str__(self):
+        return self.name
+
+
+class DirectTrackedHousePet(DirectTrackedPet):
+    trained = models.BooleanField(default=False)
+
+
 # Test custom through model to model with custom PK
 
 
 class TaggedCustomPKFood(TaggedItemBase):
     content_object = models.ForeignKey("DirectCustomPKFood", on_delete=models.CASCADE)
 
+    class Meta:
+        unique_together = [["content_object", "tag"]]
+
 
 class TaggedCustomPKPet(TaggedItemBase):
     content_object = models.ForeignKey("DirectCustomPKPet", on_delete=models.CASCADE)
 
+    class Meta:
+        unique_together = [["content_object", "tag"]]
+
 
-@python_2_unicode_compatible
 class DirectCustomPKFood(models.Model):
     name = models.CharField(max_length=50, primary_key=True)
     tags = TaggableManager(through=TaggedCustomPKFood)
@@ -120,7 +206,6 @@ class DirectCustomPKFood(models.Model):
         return self.name
 
 
-@python_2_unicode_compatible
 class DirectCustomPKPet(models.Model):
     name = models.CharField(max_length=50, primary_key=True)
 
@@ -138,8 +223,10 @@ class DirectCustomPKHousePet(DirectCustomPKPet):
 class TaggedCustomPK(CommonGenericTaggedItemBase, TaggedItemBase):
     object_id = models.CharField(max_length=50, verbose_name="Object id", db_index=True)
 
+    class Meta:
+        unique_together = [["object_id", "tag"]]
+
 
-@python_2_unicode_compatible
 class CustomPKFood(models.Model):
     name = models.CharField(max_length=50, primary_key=True)
 
@@ -149,7 +236,6 @@ class CustomPKFood(models.Model):
         return self.name
 
 
-@python_2_unicode_compatible
 class CustomPKPet(models.Model):
     name = models.CharField(max_length=50, primary_key=True)
 
@@ -174,9 +260,12 @@ class OfficialThroughModel(GenericTaggedItemBase):
     tag = models.ForeignKey(
         OfficialTag, related_name="tagged_items", on_delete=models.CASCADE
     )
+    extra_field = models.CharField(max_length=10)
+
+    class Meta:
+        unique_together = [["content_type", "object_id", "tag"]]
 
 
-@python_2_unicode_compatible
 class OfficialFood(models.Model):
     name = models.CharField(max_length=50)
 
@@ -186,7 +275,6 @@ class OfficialFood(models.Model):
         return self.name
 
 
-@python_2_unicode_compatible
 class OfficialPet(models.Model):
     name = models.CharField(max_length=50)
 
@@ -215,6 +303,11 @@ class Movie(Media):
     pass
 
 
+class ProxyPhoto(Photo):
+    class Meta:
+        proxy = True
+
+
 class ArticleTag(Tag):
     class Meta:
         proxy = True
@@ -243,7 +336,7 @@ class Article(models.Model):
 
 
 class CustomManager(models.Model):
-    class Foo(object):
+    class Foo:
         def __init__(*args, **kwargs):
             pass
 
@@ -258,15 +351,39 @@ class Child(Parent):
     pass
 
 
-@python_2_unicode_compatible
 class UUIDFood(models.Model):
     id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
     name = models.CharField(max_length=50)
     tags = TaggableManager(through="UUIDTaggedItem")
 
+    created_at = models.DateTimeField(auto_now_add=True)
+
     def __str__(self):
         return self.name
 
+    class Meta:
+        # With a UUIDField pk, objects are not always ordered by creation time. So explicitly set ordering.
+        ordering = ["created_at"]
+
+
+class UUIDPet(models.Model):
+    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
+    name = models.CharField(max_length=50)
+
+    tags = TaggableManager(through="UUIDTaggedItem")
+    created_at = models.DateTimeField(auto_now_add=True)
+
+    def __str__(self):
+        return self.name
+
+    class Meta:
+        # With a UUIDField pk, objects are not always ordered by creation time. So explicitly set ordering.
+        ordering = ["created_at"]
+
+
+class UUIDHousePet(UUIDPet):
+    trained = models.BooleanField(default=False)
+
 
 class UUIDTag(TagBase):
     id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
@@ -276,3 +393,13 @@ class UUIDTaggedItem(GenericUUIDTaggedItemBase):
     tag = models.ForeignKey(
         UUIDTag, related_name="%(app_label)s_%(class)s_items", on_delete=models.CASCADE
     )
+
+    class Meta:
+        unique_together = [["content_type", "object_id", "tag"]]
+
+
+# Exists to verify system check failure.
+# tests.Name.tags: (fields.E303) Reverse query name for 'Name.tags' clashes with field name 'Tag.name'.
+# 	HINT: Rename field 'Tag.name', or add/change a related_name argument to the definition for field 'Name.tags'.
+class Name(models.Model):
+    tags = TaggableManager(related_name="a_unique_related_name")
