@@ -7,6 +7,11 @@ from django.db import IntegrityError, connection, models
 from django.test import RequestFactory, SimpleTestCase, TestCase
 from django.test.utils import override_settings
 
+from taggit.managers import TaggableManager, _TaggableManager
+from taggit.models import Tag, TaggedItem
+from taggit.utils import edit_string_for_tags, parse_tags
+from taggit.views import tagged_object_list
+
 from .forms import (
     BlankTagForm,
     CustomPKFoodForm,
@@ -34,6 +39,7 @@ from .models import (
     Food,
     HousePet,
     Movie,
+    MultiInheritanceFood,
     Name,
     OfficialFood,
     OfficialHousePet,
@@ -55,11 +61,6 @@ from .models import (
     UUIDTaggedItem,
 )
 
-from taggit.managers import TaggableManager, _TaggableManager
-from taggit.models import Tag, TaggedItem
-from taggit.utils import edit_string_for_tags, parse_tags
-from taggit.views import tagged_object_list
-
 
 class BaseTaggingTestCase(TestCase):
     def assert_tags_equal(self, qs, tags, sort=True, attr="name"):
@@ -126,14 +127,10 @@ class CustomTagCreationTestCase(TestCase):
         apple = OfficialFood.objects.create(name="apple")
 
         # let's add two official tags
-        apple.tags.add(
-            "foo", "bar", tag_kwargs={"official": True},
-        )
+        apple.tags.add("foo", "bar", tag_kwargs={"official": True})
 
         # and two unofficial ones
-        apple.tags.add(
-            "baz", "wow", tag_kwargs={"official": False},
-        )
+        apple.tags.add("baz", "wow", tag_kwargs={"official": False})
 
         # We should end up with 4 tags
         self.assertEquals(apple.tags.count(), 4)
@@ -168,6 +165,7 @@ class TagUUIDModelTestCase(TagModelTestCase):
 
 class TaggableManagerTestCase(BaseTaggingTestCase):
     food_model = Food
+    multi_inheritance_food_model = MultiInheritanceFood
     pet_model = Pet
     housepet_model = HousePet
     taggeditem_model = TaggedItem
@@ -610,6 +608,21 @@ class TaggableManagerTestCase(BaseTaggingTestCase):
             ordered=False,
         )
 
+    def test_multi_inheritance_similarity_by_tag(self):
+        """Test that pears are more similar to apples than watermelons using multi_inheritance"""
+        apple = self.multi_inheritance_food_model.objects.create(name="apple")
+        apple.tags.add("green", "juicy", "small", "sour")
+
+        pear = self.multi_inheritance_food_model.objects.create(name="pear")
+        pear.tags.add("green", "juicy", "small", "sweet")
+
+        watermelon = self.multi_inheritance_food_model.objects.create(name="watermelon")
+        watermelon.tags.add("green", "juicy", "large", "sweet")
+
+        similar_objs = apple.tags.similar_objects()
+        self.assertEqual(similar_objs, [pear, watermelon])
+        self.assertEqual([obj.similar_tags for obj in similar_objs], [3, 2])
+
     def test_similarity_by_tag(self):
         """Test that pears are more similar to apples than watermelons"""
         apple = self.food_model.objects.create(name="apple")
@@ -648,6 +661,19 @@ class TaggableManagerTestCase(BaseTaggingTestCase):
             str(self.taggeditem_model.objects.first()), "apple tagged with juicy"
         )
 
+    def test_taggeditem_through_defaults(self):
+        if self.taggeditem_model != OfficialThroughModel:
+            self.skipTest(
+                "Through default tests are only run when the tagged item model has extra_field"
+            )
+        apple = self.food_model.objects.create(name="kiwi")
+        apple.tags.add("juicy", through_defaults={"extra_field": "green"})
+
+        self.assertEqual(
+            str(self.taggeditem_model.objects.first()), "kiwi tagged with juicy"
+        )
+        self.assertEqual(self.taggeditem_model.objects.first().extra_field, "green")
+
     def test_abstract_subclasses(self):
         p = Photo.objects.create()
         p.tags.add("outdoors", "pretty")
